//#define DEBUG  //enable debugging via an I2C LCD Screen
#include <FastLED.h>
#include "RTClib.h"

#ifdef DEBUG
#include <LiquidCrystal_I2C.h>
#endif

#define NUM_LEDS 60
#define DATA_PIN 4
#define MAX_BRIGHTNESS 250
#define RING_BRIGHTNESS_FACTOR 170

#define TWINKLE_TIME 1500
#define BRIGHTNESS_PIN A2
#define BRIGTHNESS_MIN_DELTA 7
#define DIR_SWITCH_PIN A1
#define TRANSITION_SWITCH_PIN 7 //type of transition between minutes: soft or twinkle

#define PIR_INTERRUPT_PIN 2
#define AUTO_SHUTDOWN_THRESHOLD 600000  //auto led shutdown in ms

#define UPDATES_PER_SECOND 30

//set to TRUE to force setting time to this file compile time. 
//Don't forget to reflash with FALSE or the clock time will be set to wrong time at subsequents boots.
#define FORCE_SET_TIME_TO_COMPILE_TIME false 

uint8_t prevMinutes=60;

enum states { LIGHTING_UP,
              LIT,
              LIGHTING_DOWN,
              OFF };

RTC_DS3231 rtc;

volatile bool pir_event = false;
CRGB leds[NUM_LEDS];
uint8_t brightness = 0;
uint8_t lastBrightness = 0;
volatile uint32_t lastPIRDetection = 0;

#ifdef DEBUG
LiquidCrystal_I2C lcd(0x27, 16, 2);  // I2C address 0x27, 16 column and 2 rows // for debugging/logging purposes
#endif

DateTime time;
uint32_t prev_millis = 0;

enum states state;  //current clock state

void setup() {
  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(DIR_SWITCH_PIN, INPUT_PULLUP);
  pinMode(TRANSITION_SWITCH_PIN, INPUT_PULLUP);

#ifdef DEBUG
  //init LCD screen, used for debug purposes
  lcd.init();
  lcd.backlight();
  lcd.print("Starting");
  //set cursor to home position for later print() statements
  lcd.setCursor(0, 0);
  lcd.noAutoscroll();
  delay(750);
#endif
  initRTC();
  getCurrentTime();
  FastLED.addLeds<NEOPIXEL, DATA_PIN>(leds, NUM_LEDS).setCorrection(TypicalLEDStrip);
  //attach PIR sensor to interrupt pin
  attachInterrupt(digitalPinToInterrupt(PIR_INTERRUPT_PIN), onPIRDetection, RISING);

  fillLEDsFromColorwheel(leds);
  state = LIGHTING_UP;    
}

void initRTC() {  
  //init RTC if necessary
  if (!rtc.begin()) {
#ifdef DEBUG
    lcd.print("Couldn't find RTC");
#endif
    digitalWrite(LED_BUILTIN, HIGH);
    while (1) delay(1000);
  }
  if (rtc.lostPower() || FORCE_SET_TIME_TO_COMPILE_TIME) {
#ifdef DEBUG
    lcd.print("RTC lost power,setting the time!");
#endif
    // When time needs to be set on a new device, or after a power loss, the
    // following line sets the RTC to the date & time this sketch was compiled
    //set RTC time in UTC since we need to automatically apply DST when displaying the time
    DateTime tempDT = DateTime(F(__DATE__), F(__TIME__));
    uint32_t offset = adjustDstEurope(tempDT);
    tempDT = tempDT - TimeSpan(offset);
    rtc.adjust(tempDT);
    // This line sets the RTC with an explicit date & time, for example to set
    // January 21, 2014 at 3am you would call:
    // rtc.adjust(DateTime(2014, 1, 21, 3, 0, 0));
    delay(100);
  }
  //set RTC time in UTC since we need to automatically apply DST when displaying the time
  //DateTime tempDT = DateTime(F(__DATE__), F(__TIME__));
  //uint32_t offset = adjustDstEurope(tempDT);
  //tempDT = tempDT - TimeSpan(offset);
  //rtc.adjust(tempDT);
 
}

/*
* PIR sensor interrupt routine
*/
void onPIRDetection() {
  pir_event = true;  //motion detected, we set the event variable to light-up the clock
  //digitalWrite(LED_BUILTIN, HIGH);
}

void getCurrentTime(){
  //get time and adjust for European DST
  time = rtc.now(); 
  uint32_t offset = adjustDstEurope(time);
  time = time + TimeSpan(offset);
}

//return the current minute from 0 to 59 included taking into account the clock direction
uint8_t getMinute() {
  bool clockwise = digitalRead(DIR_SWITCH_PIN);
  if(clockwise){
    return time.minute();
  }else{
    return 59 - time.minute();
  }
}

//return the current hour from 0 to 11 included taking into account the clock direction
uint8_t getHour() {
  bool clockwise = digitalRead(DIR_SWITCH_PIN);
  if(clockwise){
	  return time.hour() % 12;	
  }else{
	  return 12 - (time.hour() % 12);
  }
}

//start of LED stuff

void fillLEDsFromColorwheel(CRGB *leds){
  uint8_t initialHue = 255 - int((255.0f/12) * getHour());
  fill_rainbow_circular(leds, NUM_LEDS,initialHue); 
  //dimm the ring so that the minute led shows off
  nscale8(leds, NUM_LEDS, RING_BRIGHTNESS_FACTOR);
}

void chase(){
  //save colors:
  CRGB target_leds[NUM_LEDS];
  fillLEDsFromColorwheel(target_leds);
  fract8 	amountOfOverlay = 40;
  bool clockwise = digitalRead(DIR_SWITCH_PIN);

  for(uint8_t i=0; i< NUM_LEDS * 1.5; i++){
    if(i < NUM_LEDS){
      if(clockwise){
        leds[i] = CRGB::White;    
      }else{
        leds[NUM_LEDS-1-i] = CRGB::White;
      }
    }
    for(uint8_t o=0; o< NUM_LEDS; o++){
      nblend(leds[o], target_leds[o], amountOfOverlay);          
    }    
    FastLED.show();
    FastLED.delay(1000 / UPDATES_PER_SECOND / 1.5);
  }
}


/*
* blocking twinkle animation called before each move
*/
void twinkle() {
  uint32_t start = millis();
  while (millis() - start < TWINKLE_TIME) {
    uint8_t max = random8(NUM_LEDS / 2);
    for (int i = 0; i < max; i++) {
      leds[random8(NUM_LEDS)] = CRGB::White;
      FastLED.show();
    }
    FastLED.delay(1000 / UPDATES_PER_SECOND);
    fillLEDsFromColorwheel(leds);
  }
  FastLED.show();
  delay(100);
}

/*
* fade the LEDs in and out depending on the current state, blocking call
*/
void fade() {
  while (true) {
    switch (state) {
      case LIGHTING_UP:
        if (brightness <  getPotValue()) {
          if(brightness == 0){
            getCurrentTime(); //fetch time from RTC since we may be waking up from a sleep
            fillLEDsFromColorwheel(leds);
            FastLED.show();
          }
          brightness += 1;
        } else {
          smoothLight(getMinute()); //force lighting-up of the current minute led before the next minute transition
          state = LIT;
          return;
        }
        break;

      case LIGHTING_DOWN:
        if (brightness > 0) {
          brightness -= 1;
        } else {
          state = OFF;
          return;
        }
        break;

      default:
        return;
    }
    FastLED.setBrightness(brightness);
    FastLED.show();
    FastLED.delay(1000 / UPDATES_PER_SECOND);
  }
}

void smoothLight(uint8_t posMin){
  for(int f=0; f < UPDATES_PER_SECOND; f++){
    leds[posMin].addToRGB(4);
    FastLED.show();
    FastLED.delay(1000 / UPDATES_PER_SECOND);
  }
  leds[posMin] = CRGB::White;
  FastLED.show();
}

//end of LED stuff

void handleTime(){ 
 if(prevMinutes != getMinute()){
    //uint8_t posMin = (int((float(NUM_LEDS) / 60.0f) * getMinute())) % NUM_LEDS;       
    //since we have 60 LED for 60 minutes we can simplify as this:
    uint8_t posMin = getMinute();
    bool shouldTwinkle = !digitalRead(TRANSITION_SWITCH_PIN);
    if(shouldTwinkle){
      twinkle();
      fillLEDsFromColorwheel(leds);
      //chase animation at minute=0
      if(posMin == 0){
        chase();
      }
      smoothLight(posMin);     
      
    }else{
      fillLEDsFromColorwheel(leds);
      //chase animation at minute=0
      if(posMin == 0){
        chase();
      }
      fillLEDsFromColorwheel(leds);
      smoothLight(posMin);
    }
    
 }
 prevMinutes = getMinute();
}

void handleBrightness(){
  if(state == LIT){
    FastLED.setBrightness(getPotValue());
    FastLED.show();
  }
}

void loop() {  
  fade();  //blocking fade in or out depending on current state

  handleTime();

  handleBrightness();

  //manage state to wake-up LEDs on PIR event
  if (pir_event) {
    lastPIRDetection = millis();
    pir_event = false;
    if (state == OFF) {
      state = LIGHTING_UP;
    }
  }

  //handle auto-shutdown
  if ((millis() - lastPIRDetection) > AUTO_SHUTDOWN_THRESHOLD) {
    state = LIGHTING_DOWN;
  }



#ifdef DEBUG
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.println(getHour());
    lcd.print(":");
    lcd.print(getMinute());
    lcd.print(":");
    lcd.print(time.second());
    lcd.setCursor(0, 1);

    //lcd.print("p");
    //lcd.print(getPotValue());
#endif
  FastLED.delay(1000 / UPDATES_PER_SECOND);
  getCurrentTime();  
}

/*
* 
*/
uint8_t getPotValue() {
  uint8_t val = map(analogRead(BRIGHTNESS_PIN), 0, 880, 5, MAX_BRIGHTNESS);
  if(abs(lastBrightness - val) < BRIGTHNESS_MIN_DELTA){ 
    val = lastBrightness;
  }else{
    lastBrightness = val;
  }
  return val;
}

/*
* Calculates Daylight Saving Time offset from UTC Date, for Europe
*/
int adjustDstEurope(DateTime t) {
  /*You can use the following equations to calculate when DST starts and ends.
 The divisions are integer divisions, in which remainders are discarded.
 "mod" means the remainder when doing integer division, e.g., 20 mod 7 = 6.
 That is, 20 divided by 7 is 2 and 6/7th (where six is the remainder).
 With: y = year.
        For the United States:
            Begin DST: Sunday April (2+6*y-y/4) mod 7+1
            End DST: Sunday October (31-(y*5/4+1) mod 7)
           Valid for years 1900 to 2006, though DST wasn't adopted until the 1950s-1960s. 2007 and after:
            Begin DST: Sunday March 14 - (1 + y*5/4) mod 7
            End DST: Sunday November 7 - (1 + y*5/4) mod 7;
        European Economic Community:
            Begin DST: Sunday March (31 - (5*y/4 + 4) mod 7) at 1h U.T.
            End DST: Sunday October (31 - (5*y/4 + 1) mod 7) at 1h U.T.
            Since 1996, valid through 2099
(Equations by Wei-Hwa Huang (US), and Robert H. van Gent (EC))
 
 Adjustig Time with DST Europe/France/Paris: UTC+1h in winter, UTC+2h in summer
 
 */

  // last sunday of march
  int beginDSTDate = (31 - (5 * t.year() / 4 + 4) % 7);
  //Serial.println(beginDSTDate);
  int beginDSTMonth = 3;
  //last sunday of october
  int endDSTDate = (31 - (5 * t.year() / 4 + 1) % 7);
  //Serial.println(endDSTDate);
  int endDSTMonth = 10;
  // DST is valid as:
  if (((t.month() > beginDSTMonth) && (t.month() < endDSTMonth))
      || ((t.month() == beginDSTMonth) && (t.day() > beginDSTDate))
      || ((t.month() == beginDSTMonth) && (t.day() == beginDSTDate) && (t.hour() >= 1))
      || ((t.month() == endDSTMonth) && (t.day() < endDSTDate))
      || ((t.month() == endDSTMonth) && (t.day() == endDSTDate) && (t.hour() < 1)))
    return 7200;     // DST europe = utc +2 hour (summer time)
  else return 3600;  // nonDST europe = utc +1 hour (winter time)
}